from __future__ import annotations

import bs4
import httpx
import ujson as json
from typing import Any, cast
from uuid import uuid4
from pydantic import Field
from asynctinydb import Query, Document
from datetime import datetime, UTC
from urllib.parse import urlparse
from .base import BaseCrawler
from ..classes import Book, Chapter, Episode, Line
from ..classes import TimeMeta, Context, Task, ImageRef


class KakuyomuCrawler(BaseCrawler):
    """
    Crawler for kakuyomu.jp
    """
    name: str = Field(default_factory=lambda: f"kakuyomu-{str(uuid4())[:8]}")
    description: str = "A crawler for kakuyomu.jp"

    async def crawl_stream(
            self,
            task: Task,
            ctx: Context,
    ):
        logger = ctx.logger
        db = ctx.db
        cli = ctx.client
        url = urlparse(task.url)
        url_path = url.path.removesuffix('/')
        origin = f"{url.scheme}://{url.netloc}"
        logger.info("%s: Crawling URL %s", self, task.url)
        r = await cli.get(task.url)
        r.raise_for_status()

        soup = bs4.BeautifulSoup(r.text.replace('\u3000', "  "), "lxml")
        data = json.loads(soup.find(id="__NEXT_DATA__").text)
        items = data["props"]["pageProps"]["__APOLLO_STATE__"]
        work_info: dict[str, Any] = items[f"Work:{url_path.removeprefix("/works/")}"]
        title: str = work_info["title"]
        author = items[work_info["author"]["__ref"]]["name"]

        logger.info("%s:  book %s by %s", self, title, author)
        book_query = Query()
        data: Document = await db.get(
            book_query.title == title and book_query.author == author)
        if data:
            logger.debug("%s: Book %s retrieved from database", self, title)
            book = Book.model_validate(dict(data))
        else:
            book = Book(title=title, author=author)

        book.url = task.url

        cover = work_info["adminCoverImageUrl"]
        if cover is None:
            logger.info("%s: No cover found for %s", self, title)
        else:
            book.cover = ImageRef(url=cover["content"])

        desc = work_info.get("catchphrase", '')
        desc += f"\n{work_info["introduction"]}"
        book.description = desc

        book.time_meta.created_at = datetime.fromisoformat(
            work_info["publishedAt"]).astimezone(UTC)
        book.time_meta.updated_at = datetime.fromisoformat(
            work_info.get("lastEpisodePublishedAt",
                          book.time_meta.created_at)).astimezone(UTC)
        book.tags.update(work_info["tagLabels"])
        if work_info["isCruel"]:
            book.tags.add("残酷描写")
        if work_info["isViolent"]:
            book.tags.add("暴力描写")
        if work_info["isSexual"]:
            book.tags.add("性描写")

        toc = [items[ref["__ref"]] for ref in work_info["tableOfContents"]]

        for chapter_info in toc:
            chapter_name = items[chapter_info["chapter"]["__ref"]]["title"]

            logger.info("%s: Crawling chapter %s", self, chapter_name)
            old_chapter = book.get_chapter(chapter_name)
            if old_chapter is None:
                logger.debug("%s: New chapter %s created", self, chapter_name)
                chapter = Chapter(title=chapter_name)
                book.chapters.append(chapter)
            else:
                logger.debug("%s: Chapter %s retrieved from database",
                             self, chapter_name)
                chapter = old_chapter

            for episode_info in (items[ep["__ref"]] for ep in chapter_info["episodeUnions"]):

                episode_url = f"{url.scheme}://{url.netloc}{url_path}/episodes/{episode_info["id"]}"

                subtitle = episode_info["title"]
                logger.info("%s: Crawling episode %s from %s",
                            self, subtitle, episode_url)

                published_at = datetime.fromisoformat(
                    episode_info["publishedAt"]).astimezone(UTC)
                created_at = published_at

                episode = chapter.get_episode(subtitle)
                if episode is not None:
                    logger.debug("%s: Episode %s retrieved from database",
                                 self, subtitle)
                    if episode.time_meta.created_at is None:
                        episode.time_meta.created_at = published_at
                    episode.time_meta.updated_at = published_at
                    created_at = episode.time_meta.created_at

                    if episode.time_meta.saved_at < published_at:
                        logger.debug("%s: Episode %s updated", self, subtitle)
                        episode_index = chapter.episodes.index(episode)
                        episode = await self.get_episode(
                            episode_url, cli,
                            created_at=created_at, updated_at=published_at)
                        chapter.episodes[episode_index] = episode
                else:
                    logger.debug("%s: New episode %s created", self, subtitle)
                    episode = await self.get_episode(
                        episode_url, cli,
                        created_at=created_at, updated_at=published_at)
                    chapter.episodes.append(episode)

                if not episode.lines:
                    logger.warning("%s: Episode %s has no content", self, subtitle)

                await db.upsert(book.model_dump(mode="json"),
                                book_query.title == title and book_query.author == author)

                yield task, book, chapter, episode

        await db.upsert(book.model_dump(mode="json"),
                        book_query.title == title and book_query.author == author)

    async def get_episode(
            self,
            url: str,
            client: httpx.AsyncClient,
            *,
            created_at: datetime | None = None,
            updated_at: datetime | None = None,
    ) -> Episode:
        r = await client.get(url)
        r.raise_for_status()

        if created_at is not None:
            created_at = created_at.astimezone(UTC)
        if updated_at is not None:
            updated_at = updated_at.astimezone(UTC)

        soup = bs4.BeautifulSoup(r.text.replace('\u3000', "  "), "lxml")
        body = soup.body
        title = body.find(class_="widget-episodeTitle").text.strip()
        honbun = body.find(class_="widget-episodeBody")
        lines = list[Line]()
        for p in honbun.find_all("p"):
            p: bs4.Tag
            for br in p.find_all("br"):
                br: bs4.Tag
                br.replace_with(f"\n{br.decode_contents()}")
            decoded = str(p.decode_contents())
            if not decoded.strip():
                decoded = ''  # replace empty lines with empty string
            lines.append(Line(content=decoded))

        episode = Episode(
            title=title,
            lines=lines,
            time_meta=TimeMeta(
                created_at=created_at, updated_at=updated_at),
        )

        return episode



from __future__ import annotations

import ujson as json
import asyncio
from random import randint
from typing import cast, overload
from uuid import uuid4
from pydantic import Field
from asynctinydb import Query
from gptbot import Bot, Message, Role, Model
from ..misc import LANG_NAME_TABLE

from bookbaker.classes import Context, Task
from ..classes import Book, Chapter, Episode
from .base import BaseTranslator


class GPTTranslator(BaseTranslator):
    """
    # GPTTranslator Class
    """
    name: str = Field(default_factory=lambda: f"gpt-{str(uuid4())[:8]}")
    description: str = "GPT Translator"
    max_retries: int | None = 10
    """Max retries before giving up"""
    batch_size: int = 512
    """Max characters to send in one batch"""
    max_tokens: int | None = 5000
    """Max tokens to preserve"""
    remind_interval: int | None = 10
    """Interval to remind glossaries"""
    skip_translated: bool = True
    """Skip already translated lines"""
    backend: Bot = Field(
        default_factory=lambda: Bot(model=Model.GPT4TurboPreview, api_key=''))

    async def translate(
            self,
            episode: Episode,
            task: Task,
            ctx: Context,
            chapter: Chapter | None = None,
            book: Book | None = None,
    ) -> Episode:

        if episode.fully_translated and self.skip_translated:
            return episode

        cli = ctx.client
        self.backend._cli = cli
        db = ctx.db
        sauce_lang = LANG_NAME_TABLE.get(task.sauce_lang.upper(), task.sauce_lang)
        target_lang = LANG_NAME_TABLE.get(task.target_lang.upper(), task.target_lang)
        logger = ctx.logger
        sess = self.backend.new_session()

        prompt = ("You are an professional translator. "
                  f"Translate JSON values from {sauce_lang} to {target_lang}.\n"
                  "Add the missing subject to the sentence\.\n"
                  "You are allowed to rephrase them to make them more natural and correct errors in original content.\n"
                  "For JSON, you should output exact the same structure as input\n"
                  "For pure text, you should output exact the same line count as input and do not modify '\\n' symbol\n"
                  #   "If the input is a list, the output order should be the same\n"
                  )
        if task.glossaries:
            prompt += "Translation reference to follow, you will be constantly reminded:\n"
            prompt += '\n'.join(f"{k} : {v}" for k, v in task.glossaries)
        if book is not None:
            prompt += "\nThe book you are translating:\n"
            book_full_meta = {
                "title": book.title,
                "description": book.description,
                "series": book.series,
                "tags": list(book.tags),
            }
            prompt += f"{json.dumps(book_full_meta, ensure_ascii=False)}\n"
        if chapter is not None:
            prompt += "\nThe chapter you are translating:\n"
            chapter_meta: dict[str, str | None] = {
                "title": chapter.title,
            }
            prompt += f"{json.dumps(chapter_meta, ensure_ascii=False)}\n"

        prompt += "\nThe episode you are translating:\n"
        episode_meta: dict[str, str | None] = {
            "title": episode.title,
            "notes": episode.notes,
        }
        prompt += f"{json.dumps(episode_meta, ensure_ascii=False)}\n"
        prompt += "\nYou can start translating now:"

        logger.debug("%s: generated prompt: %s", self, prompt)

        self.backend.prompt = prompt

        def remind():
            if task.glossaries:
                logger.debug("%s: Sending reminder", self)
                sess.append(
                    Message(role=Role.User, content=f"[{", ".join(g[0] for g in task.glossaries)}]"))
                sess.append(
                    Message(role=Role.Assistant, content=f"[{", ".join(g[1] for g in task.glossaries)}]"))
        remind()
        cycle = 0

        @overload
        async def translate(c: dict[str, str | None]) -> dict[str, str | None]:
            ...

        @overload
        async def translate(c: list[str]) -> list[str]:
            ...

        async def translate(
                c: list[str] | dict[str, str | None]
        ) -> list[str] | dict[str, str | None]:
            nonlocal cycle
            if isinstance(c, dict):
                pstr = json.dumps(c, ensure_ascii=False)
            else:
                pstr = '\n'.join(map(lambda x: x.replace('\n', r"\n"), c))
            logger.debug("%s: Sending prompt: %s", self, pstr)
            sess.trim(self.max_tokens)
            sess_bak = sess.messages.copy()

            retry = 0

            while True:
                self.backend.seed = randint(0, 10000)
                # resp: str | None = None
                # try:
                #     resp = await sess.send(jstr, ensure_json=True)
                #     logger.debug("%s: Received response: %s", self, resp)
                #     obj = json.loads(cast(str, resp))
                #     keys = set(c.keys()) if isinstance(c, dict) else set(("list",))
                #     if set(obj.keys()) != keys:
                #         raise ValueError("Key mismatch")
                #     if isinstance(c, list):
                #         if len(obj["list"]) != len(c):
                #             raise ValueError("Length mismatch")
                #         return obj["list"]

                #     return obj
                resp: str = "None"
                try:
                    resp = await sess.send(pstr, ensure_json=False)
                    logger.debug("%s: Received response: %s", self, resp)
                    if isinstance(c, list):
                        obj = resp.split('\n')
                        obj = list(
                            filter(None, map(lambda x: x.replace(r"\n", '\n'), obj)))
                        if not isinstance(obj, list):
                            raise TypeError("Expected list, got dict")
                        if len(obj) != len(c):
                            raise ValueError("Length mismatch")
                    else:
                        obj = json.loads(resp)
                        if not isinstance(obj, dict):
                            raise TypeError("Expected dict, got list")
                        if obj.keys() != c.keys():
                            raise ValueError("Key mismatch")

                    return obj

                except Exception as e:
                    logger.debug("%s: Failed to get valid response: %s", self, resp)
                    logger.exception(e)
                    sess.messages = sess_bak
                    retry += 1
                    if self.max_retries is not None and retry > self.max_retries:
                        raise RuntimeError(
                            "Failed to get valid response in %d retries" % self.max_retries)
                    await asyncio.sleep(1)
                    continue

                finally:
                    cycle += 1
                    if self.remind_interval is not None and cycle >= self.remind_interval:
                        remind()
                        cycle = 0

        if book is not None and None in (
            book.title_translated,
            book.description_translated,
            book.series_translated if book.series else ''
        ):
            book_meta: dict[str, str | None] = {
                "title": book.title,
                "description": book.description,
                "series": book.series
            }
            book_meta = await translate(book_meta)
            book.title_translated = book_meta["title"]
            book.description_translated = book_meta["description"]
            book.series_translated = book_meta["series"]

        if chapter is not None and chapter.title_translated is None:
            chapter_meta = {
                "title": chapter.title
            }
            chapter_meta = await translate(chapter_meta)
            chapter.title_translated = chapter_meta["title"]

        episode_meta = await translate(episode_meta)
        episode.title_translated = episode_meta["title"]
        episode.notes_translated = episode_meta["notes"]

        indexes = list[int]()
        contents = list[str]()
        c_cnt = 0
        MAX_CNT = self.batch_size

        for i, line in enumerate(episode.lines):
            if not line.content.strip():
                continue
            if self.skip_translated and line.translated is not None:
                continue
            indexes.append(i)
            contents.append(line.content)
            c_cnt += len(line.content)
            if c_cnt > MAX_CNT:
                translated = await translate(contents)
                for j, v in zip(indexes, translated):
                    o = episode.lines[j].content
                    line = episode.lines[j]
                    line.translated = v
                    line.candidates[self.name] = v
                    if not v.strip():
                        logger.warning("%s: Empty translation for %d", self, o)
                indexes.clear()
                contents.clear()
                c_cnt = 0

        if indexes:
            translated = await translate(contents)
            for j, v in zip(indexes, translated):
                line = episode.lines[j]
                line.translated = v
                line.candidates[self.name] = v
                for j, v in zip(indexes, translated):
                    o = episode.lines[j].content
                    line = episode.lines[j]
                    line.translated = v
                    line.candidates[self.name] = v
                    if not v.strip():
                        logger.warning("%s: Empty translation for %s", self, o)

        q = Query()
        if book is not None:
            await db.upsert(book.model_dump(mode="json"),
                            q.title == book.title and q.author == book.author)

        return episode



from base64 import b64decode
from vermils.io import aio
from .classes import Context


async def get_url_content(url: str, origin: str, ctx: Context) -> bytes:
    if url.startswith("http"):
        r = await ctx.client.get(
            url,
            headers={"Origin": origin},
            follow_redirects=True,
        )
        r.raise_for_status()
        data = r.content
    elif url.startswith("base64://"):
        data = b64decode(url.removeprefix("base64://"))
    elif url.startswith("file://"):
        async with aio.open(url.removeprefix("file://"), "rb") as f:
            data = await f.read()
    else:
        raise ValueError(f"Cannot download URL: {url}")
    return data



from __future__ import annotations

import httpx
import urllib.request
import logging
from typing import Literal, Any
from pydantic import BaseModel, Field, ConfigDict
from pathlib import Path
from asynctinydb import TinyDB, CachingMiddleware, JSONStorage
from asynctinydb.table import StrID
from vermils.gadgets.monologger import MonoLogger
from .classes import Task


class BaseConfig(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        populate_by_name=True,
        extra="ignore"
    )


class Config(BaseConfig):
    """
    # Config Class
    """
    class ClientConfig(BaseConfig):
        """
        # ClientConfig Class
        """
        user_agent: str = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36"
        timeout: float = 10.0
        proxy: str | None = None
        max_retries: int = 5
        trust_env: bool = True

        def get_cli(self) -> httpx.AsyncClient:
            """
            # Get Client
            """
            proxy = self.proxy
            if self.proxy is None and self.trust_env:
                sys_proxy = urllib.request.getproxies()
                proxy = sys_proxy.get('https', sys_proxy.get('http'))
            trans = httpx.AsyncHTTPTransport(
                http2=True,
                trust_env=False,
                proxy=proxy,
                retries=self.max_retries,
            )
            return httpx.AsyncClient(
                headers={"User-Agent": self.user_agent},
                transport=trans,
                timeout=self.timeout)

    class DBConfig(BaseConfig):
        """
        # DBConfig Class
        """
        path: Path = Path('db.json')
        write_buffer_size: int = 0

        def get_db(self, dir_path: Path | None = None) -> TinyDB:
            """
            # Get DB
            """
            if self.write_buffer_size:
                storage: Any = CachingMiddleware(
                    JSONStorage, self.write_buffer_size
                )
            else:
                storage = JSONStorage
            path = dir_path / self.path if dir_path else self.path
            return TinyDB(
                str(path),
                storage=storage,
                indent=2,
                ensure_ascii=False,
            )

    class LoggingConfig(BaseConfig):
        """
        # LoggingConfig Class
        """
        level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO"
        fmt: str = '%(asctime)s - %(levelname)s - %(message)s'
        dirpath: Path = Path('logs')
        show_console: bool = True

        def get_logger(self, name: str = "root"):
            """
            # Get Logger
            """
            logger = MonoLogger(
                name=name,
                level=self.level,
                path=str(self.dirpath),
                formatter=self.fmt,
            )
            if self.show_console:
                handler = logging.StreamHandler()
                handler.setFormatter(logging.Formatter(self.fmt))
                logger.addHandler(handler)
            return logger

    version: str = "1.0.0"
    tasks: list[Task] = Field(default_factory=list)
    roles: list[Any] = Field(default_factory=list)
    client: ClientConfig = Field(default_factory=ClientConfig)
    db: DBConfig = Field(default_factory=DBConfig)
    logging: LoggingConfig = Field(default_factory=LoggingConfig)



LANG_NAME_TABLE = {
    "AR": "Arabic",
    "BG": "Bulgarian",
    "CS": "Czech",
    "DA": "Danish",
    "DE": "German",
    "EL": "Greek",
    "EN": "English",
    "EN-GB": "English (British)",
    "EN-US": "English (American)",
    "ES": "Spanish",
    "ET": "Estonian",
    "FI": "Finnish",
    "FR": "French",
    "HU": "Hungarian",
    "ID": "Indonesian",
    "IT": "Italian",
    "JA": "Japanese",
    "KO": "Korean",
    "LT": "Lithuanian",
    "LV": "Latvian",
    "NB": "Norwegian (Bokmål)",
    "NL": "Dutch",
    "PL": "Polish",
    "PT": "Portuguese",
    "PT-BR": "Portuguese (Brazilian)",
    "PT-PT": "Portuguese (Portugal)",
    "RO": "Romanian",
    "RU": "Russian",
    "SK": "Slovak",
    "SL": "Slovenian",
    "SV": "Swedish",
    "TR": "Turkish",
    "UK": "Ukrainian",
    "ZH": "Chinese (Simplified)",
}
